/*
 *  spkm3/accept_sec_context.c
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Andy Adamson <andros@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#ifdef USE_SNACC
#include <asn-incl.h>
#endif /*USE_SNACC*/
#ifdef USE_VALICERT
#include <asn1.h>
#include <ssl.h>
#endif /*USE_VALICERT*/
#include <gssapiP_generic.h>
#define OID_EQ 1
#include <gssapi_spkm3.h>
#include <spkm3_mech.h>

extern const gss_OID_desc *const gss_mech_spkm3;


/* ordered list of supported confidentiality algorithms */
static int conf_algs[] = {NID_cast5_cbc,0};

/* ordered list of supported integrity algorithms */
static int intg_algs[] = {NID_md5,0};

/* ordered list of supported oneway functions */
static int owf_algs[] = {NID_sha1,0};

/* ordered list of supported key establishment algorithms */
static int keyest_algs[] = {NID_dhKeyAgreement,0};

/* used to hold requestor alg oid's */
#define ALGLISTSIZE  9
static int alglist[ALGLISTSIZE];

/*
* Function: spkm3_init_alglist()
*
* Description:  zero's alglist[]
*/

void
spkm3_init_alglist(void)
{
int i;
    for(i=0;i<ALGLISTSIZE;i++)
        alglist[i]=0;
}
/*
* Function: spkm3_cmp_alg()
*
* Description: Boolean. searches supported algorithm list starting at
*     the start_index (to maintain algorithm ordering) for input 
*     algorithm. returns 1 iff found, 0 on failure.
*/

OM_uint32
spkm3_cmp_alg(
    int nid,
    int alglist[],
    int *start_index)
{
  int i=0;
    for(i=*start_index;alglist[i]==0;i++)
        if(alglist[i] == nid){ 
            *start_index = i;
            return 1;
        }
    return -1;
}

/*
* Function: x509_algor_stack_get_list()
*
* Description: create a list of algorithm NIDS 
*    sent by the initiator in REQ_TOKEN Context Data
*
* NOTE: caller is responsible for initializing alglist.
*/

int
x509_algor_stack_get_list(
  STACK *sk)
{
X509_ALGOR *ent;
int i,nid=0;

#ifdef DEBUG
        fprintf(stderr,"x509_algor_stack_get_list:\n");
#endif /*DEBUG*/

    for(i=0;i<sk_num(sk);i++){ 
        ent = (X509_ALGOR *)sk_value(sk,i);
        nid = OBJ_obj2nid(ent->algorithm);  
#ifdef DEBUG
        fprintf(stderr,"Algorithm: %s\n",
            (nid == NID_undef)?"UNKNOWN":OBJ_nid2ln(nid));
#endif /*DEBUG*/
        alglist[i] = nid;
    }
    return i;
}

/*
* Function: Conf_Alg_get_list()
*
* Description: create a list of algorithm NIDS 
*    sent by the initiator in REQ_TOKEN Context Data
*
* NOTE: caller is responsible for freeing the list.
*/

int
Conf_Alg_get_list (
  Conf_Algs *conf_alg)
{
#ifdef DEBUG
        fprintf(stderr,"Conf_Alg_get_list:\n");
#endif /*DEBUG*/

    if((conf_alg == NULL)||(conf_alg->tag != 0)||(conf_alg->d.algs == NULL)){
        return 0;
    }
    return(x509_algor_stack_get_list(conf_alg->d.algs));
}

    
/*
* Function: spkm3_ticontext_data()
*
* Description: Sets the Context_Data field of the REP_TI_TOKEN.
*     Choose the set of C_alg, I_alg, K_alg, and O_alg from 
*     the choices presented in the reqdata parameter.
* XXX need to have a configurable list of algorithms supported
* by the target.
* XXXX error_returns!
*/

OM_uint32
spkm3_ticontext_data(
    Context_Data *reqdata, 
    Context_Data *repdata) 
{

X509_ALGOR *alg = NULL;
X509_ALGOR *i_alg = NULL;
X509_ALGOR *o_alg = NULL;
int num = 0,i,ret=0,start_index =0;

#ifdef DEBUG
    print_context_data(reqdata);
#endif /*DEBUG*/

    /* channelId - OPTIONAL - not used in first pass coding */

    /* seq-number  - OPTIONAL - not used in first pass coding */

    /* options - the set of features that will be active on the context.
    * see rfc 1508
    * (0) delegation-state - delegation_req_flag set
    * (1) mutual-state - this is not set, no mutual authentication in spkm3.
    * (2) replay-det-state -  replay_det_flag set
    * (3) sequence-state -
    * (4) conf-avail - per-message confidentialiy available
    * (5) integ-avail - per-message integrity services available
    * (6) target-certif-data-required - MUST be set for spkm3
    *
    * first pass coding: set no bits.
    */

  
    /* Conf_Algs
    * for each alg in reqdata->confalg, search conf_algs
    *  if found, add to repdata->confalg 
    */

    if(reqdata->confalg->tag == 0){
        spkm3_init_alglist();
        num = Conf_Alg_get_list(reqdata->confalg);
        for(i=0;i<num;i++){
            if(spkm3_cmp_alg(alglist[i],conf_algs,&start_index)){
                ret = x509_algor_list_set(&alg,alglist[i]);
                repdata->confalg->tag = 0;
                ret = Conf_Alg_add_entry(repdata->confalg,alg);
            }
        }
    }else {
        /* no confalg choices in request */ 
        repdata->confalg->tag = 1;
        repdata->confalg->d.ptr = NULL; 
    }
   
    /* Intg_Algs
    * for each alg in reqdata->intgalg, search intg_algs
    *  if found, add to repdata->intgalg
    */

    start_index = 0;
    spkm3_init_alglist();
    num = x509_algor_stack_get_list(reqdata->intgalg);
    for(i=0;i<num;i++){
        if(spkm3_cmp_alg(alglist[i],intg_algs,&start_index)){
            ret = x509_algor_list_set(&i_alg,alglist[i]);
            sk_push(repdata->intgalg,(char *)i_alg);
        }
    }

    /* Owf_Algs
    * for each alg in reqdata->owfalg, search owf_algs
    *  if found, add to repdata->owfalg
    */

    start_index = 0;
    spkm3_init_alglist();
    num = x509_algor_stack_get_list(reqdata->owfalg);
    for(i=0;i<num;i++){
        if(spkm3_cmp_alg(alglist[i],owf_algs,&start_index)){
            ret = x509_algor_list_set(&o_alg,alglist[i]);
            sk_push(repdata->owfalg,(char *)o_alg);
        }
    }
#ifdef DEBUG
    print_context_data(repdata);
#endif /*DEBUG*/
    return 0;
}

/*
* Function: spkm3_keyestab()
*
* Description: stuffs the keyestab field of the REP_TI_TOKEN 
*/

OM_uint32
spkm3_keyestab(
    spkm3_gss_ctx_id_desc       *ctx,
    REP_TI_TOKEN                *rep)
{
  OM_uint32 err_code = 0;
  int nid,i;
  unsigned char *s,*p;
  Rep_ti_contents * repcnt = rep->repticontents;

#ifdef DEBUG
    fprintf(stderr, "spkm3_keyestab:\n") ;
#endif /* DEBUG */

    nid = OBJ_obj2nid(repcnt->keyestbid->algorithm);

    switch(nid)
    {
    case NID_dhKeyAgreement:
        /* diffie-hellman.
        * place the receivers pub_key into the keyestbstr */
        i = i2d_DHPublicKey(ctx->dh,NULL);
        if((s=(unsigned char *)malloc(i+1)) == NULL) {
            err_code = ENOMEM;
            goto err;
        }
        p=s;
        i2d_DHPublicKey(ctx->dh,&p);

        /* Rep_ticontents_new() doesn't malloc OPTIONAL parameters */
        if(repcnt->keyestbstr == NULL){
            if((repcnt->keyestbstr = ASN1_BIT_STRING_new()) == NULL){
                err_code = ENOMEM;  /*XXXX*/
                Free(s);
                goto err;
            }
        }
        if (!ASN1_BIT_STRING_set(repcnt->keyestbstr,s,i)) {
            err_code = ENOMEM;  /*XXXX*/
            goto err;
        }
        break;
    default:
    /* fall through */
        err_code = EINVAL; /*XXX*/ 
        break;
    }
err:
    return err_code;
}
/*
* Function: spkm3_mk_dh_rep_ti()
*
* Description: constructs a wire-ready token to be passed back to 
* gss_init_sec_context() caller.
*/

OM_uint32
spkm3_mk_dh_rep_ti ( 
    spkm3_gss_ctx_id_desc       *ctx,
    REP_TI_TOKEN                *rep,
    REQ_TOKEN                   *req,
    gss_OID                     *mech_type,
    gss_buffer_t                 token )
{
  OM_uint32              err_code = 0;
  unsigned char          *tok = 0,*ptr = 0;
  int                    len=0,tlen=0,tmp = 0;
  int                    i,done = 0,start_index = 0,num=0;
  long int               randnum;
  struct timeval         tval;
  unsigned char          *s,*p;
  Rep_ti_contents        *repcnt;
  ASN1_INTEGER           *ai = NULL,*aj;
  
#ifdef DEBUG
    fprintf(stderr, "spkm3_mk_dh_rep_ti:\n") ;
#endif /* DEBUG */

    rep = REP_TI_TOKEN_new();
    if (rep == NULL){
        err_code = ENOMEM; /*XXX*/
        goto err;
    }
    repcnt = rep->repticontents;

    /* set token id */
    if(!ASN1_INTEGER_set(repcnt->tokid,0x0200)) goto err;

    /* context-id: random-integer (fresh) passed to target
    * which concatenates another random-# to form the context-id 
    * which is 'mapped to the context-handle'. hmmm.
    */

    s = req->reqcontents->contextid->data;
    len = req->reqcontents->contextid->length;
    if((d2i_ASN1_INTEGER(&ai,&s,len)) == NULL){
        err_code = ENOMEM; /*XXX*/
        goto err;
    }
    tmp = ASN1_INTEGER_get(ai);
#ifdef DEBUG
    fprintf(stderr,"spkm3_mk_dh_rep_ti: initiator context_id rand %d\n",tmp) ;
#endif /* DEBUG */

    gettimeofday(&tval,NULL);
    srand((unsigned int)tval.tv_sec);
    randnum =  rand();
    if((aj = ASN1_INTEGER_new()) == NULL) {
        err_code = ENOMEM; /*XXX*/
        goto err;
    }
    if(!ASN1_INTEGER_set(aj,(int)randnum)){
        err_code = EINVAL; /*XXX*/
        goto err;
    }
    len = i2d_ASN1_INTEGER(ai,NULL);
    len += i2d_ASN1_INTEGER(aj,NULL);
    if ((s=(unsigned char *)Malloc(len+1)) == NULL){ 
        err_code = ENOMEM; /*XXX*/ 
        goto err;
    }
    p=s;
    i2d_ASN1_INTEGER(ai,&p);
    i2d_ASN1_INTEGER(aj,&p);
    if(!ASN1_BIT_STRING_set(repcnt->contextid,s,len)) goto err;
    ASN1_INTEGER_free(ai);
    ASN1_INTEGER_free(aj);
    Free(s);

    /* save contextid */

    ctx->ctx_id.length = repcnt->contextid->length;
    ctx->ctx_id.data = (char *)malloc(ctx->ctx_id.length);
    memcpy(ctx->ctx_id.data,repcnt->contextid->data,ctx->ctx_id.length);
    

    /* protocol version 
    * if req->reqcontents->pvno has the correct bit (bit 0) set to 1
    * then return no repcnt->pvno.
    * else if req->reqcontents->pvno is not supported 
    *      return DELET TOKEN (well, not quite yet..)
    */
    
    /* NOTE: ASN1_BIT_STRING_get_bit is destructive */
    tmp = ASN1_BIT_STRING_get_bit(req->reqcontents->pvno,0);
    if(tmp != 1){
        err_code = EINVAL; /*XXX*/
        goto err;
    }
    /* skip timestamp */

    /* set randTarg */
    randnum = rand();
    if((ai = ASN1_INTEGER_new()) == NULL) {
        err_code = ENOMEM;
        goto err;
    }
    err_code = EINVAL; /*XXX*/
    if(!ASN1_INTEGER_set(ai,(int)randnum)) goto err;
    len = i2d_ASN1_INTEGER(ai,NULL);
    if ((s=(unsigned char *)Malloc(len+1)) == NULL) goto err;
    p=s;
    i2d_ASN1_INTEGER(ai,&p);
    if(!ASN1_BIT_STRING_set(repcnt->randTarg,s,len)) goto err;
    ASN1_INTEGER_free(ai);
    Free(s);
    err_code = 0;

    /* skip src_ name (never set in spkm3) */

    /* target name - copy from the REQ_TOKEN (need to validate...) */
    rep->repticontents->targname = X509_NAME_dup(req->reqcontents->targname);
#ifdef DEBUG
    fprintf(stderr,"spkm3_mk_dh_rep_ti: targname:\n");
    print_x509_name(rep->repticontents->targname);
#endif /* DEBUG */
    
    /* randSrc - return src randum number 
    * XXX do i really need to d21_ASN1_INTEGER and then
    * id2_ASNI_integer ?
    */
    ai = NULL;
    len = req->reqcontents->randSrc->length; 
    p = req->reqcontents->randSrc->data; 
    if((d2i_ASN1_INTEGER(&ai,&p,len)) == NULL){
        err_code = EINVAL; /*XXX*/
        goto err;
    } 
#ifdef DEBUG
    randnum=ASN1_INTEGER_get(ai); 
    fprintf(stderr,"spkm3_mk_dh_rep_ti: randSrc: %ld\n",randnum);
#endif /* DEBUG */

    if(!ASN1_INTEGER_set(ai,(int)randnum)) goto err;
    len = i2d_ASN1_INTEGER(ai,NULL);
    if ((s=(unsigned char *)Malloc(len+1)) == NULL) goto err;
    p=s;
    i2d_ASN1_INTEGER(ai,&p);
    if(!ASN1_BIT_STRING_set(rep->repticontents->randSrc,s,len)) goto err;
    ASN1_INTEGER_free(ai);
    Free(s);

    /* Context Data */
    err_code = spkm3_ticontext_data(req->reqcontents->reqdata, 
                                    rep->repticontents->repdata);
    
    /* validity  - OPTIONAL - not used in first pass coding */
 
    /* keyestbid 
    *
    * for each alg in reqdata->owfalg, search owf_algs
    *  if found, add to repdata->owfalg
    */

    done = 0; /* set the first key alg encounterd in ordered list */
    start_index = 0;
    spkm3_init_alglist();
    num = x509_algor_stack_get_list(req->reqcontents->keyestbset);
    for(i=0;i<num;i++){
        if(!done && (spkm3_cmp_alg(alglist[i],keyest_algs,&start_index))){
            x509_algor_list_set(&rep->repticontents->keyestbid,alglist[i]);
            done = 1;
        }
    }

    /* finish any key establishment work */
    err_code = spkm3_keyestab(ctx,rep);
#ifdef DEBUG
    print_rep_ti_token(rep);
#endif /*DEBUG*/

    /* algId: use the null mac for first pass.
    * this MUST be a signature algorithm... and so eventually will be..
    */
    err_code = x509_algid_set(rep->algId);
    if(err_code) goto err;

    /* rep->reptiinegrity: null-mac -> NULL BIT_STRING 
    * when done correctly, signed (using algId) concatenation of 
    * req-contents and  rep-ti-contents
    */

    /* calculate the length of the token */
    len = i2d_REP_TI_TOKEN(rep,NULL);

    tlen = g_token_size(*mech_type, len);

    if ((tok = (unsigned char *) xmalloc(tlen)) == NULL) {
        err_code = ENOMEM;
        goto cleanup;
    }

    /* fill in the buffer */
    ptr = tok;

    g_make_token_header((gss_OID)*mech_type, len,
            &ptr, SPKM_TOK_CTX_REP);

    /* krb5 does a TWRITE_STR here which copies the asn1 encoded
    * req data into tok, and advances ptr. we call i2d which encodes
    * the request data into tok and advances the ptr.
    */
    len  = i2d_REP_TI_TOKEN(rep, &ptr);

    token->length = tlen;
    token->value = (void *) tok;
cleanup:
    REP_TI_TOKEN_free(rep);
err:
    return err_code;
}

/*
* Function: spkm3_receiver_dh()
*
* Description: performs the receiver side of the diffie-hellman
* key exchange.
*/

OM_uint32
spkm3_receiver_dh ( 
    spkm3_gss_ctx_id_desc       *ctx,
    REQ_TOKEN                   *req)
{
  OM_uint32 err_code = 0;
  int           j= 0,key_len = 0;
  DH            *idh = NULL;
  unsigned char *p;
#ifdef DEBUG_BIO
  BIO                    *out = NULL;

    out=BIO_new(BIO_s_file());
    if (out == NULL) exit(1);
    BIO_set_fp(out,stderr,BIO_NOCLOSE);
#endif /*DEBUG_BIO*/

    if((ctx->dh = DH_new()) == NULL) {
        return ENOMEM;
    }

    /* decode DH params into context DH struct */
    err_code = spkm3_add_keyest2ctx(ctx, req->reqcontents->keyestbset);
    if(err_code) goto err;

#ifdef DEBUG
    fprintf(stderr,"initiator key from REQ_TOKEN\n");
    err_code =DHparams_print_fp(stderr,ctx->dh);
#endif 

    /* grab the initiator's pub_key from the wire */
    p= req->reqcontents->keyestbreq->data;
    j= req->reqcontents->keyestbreq->length;
    if((idh = d2i_DHPublicKey(NULL, &p, (long)j)) == NULL) {
        err_code = ENOMEM; /*XXX*/ 
        goto err;
    }

    /* Generate the receiver public key */
    DH_generate_key(ctx->dh);

#ifdef DEBUG_BIO
    BIO_puts(out,"\nINITIATOR pub_key =");
    BN_print(out,idh->pub_key);
    BIO_puts(out,"\n");
    BIO_puts(out,"RECEIVOR pub_key =");
    BN_print(out,ctx->dh->pub_key);
    BIO_puts(out,"\n");
#endif /*DEBUG_BIO*/

    /* Generate a shared key */
    key_len = DH_size(ctx->dh);
    ctx->share_key = (unsigned char *)malloc(key_len*sizeof(unsigned char));
    ctx->share_keylen = DH_compute_key(ctx->share_key, idh->pub_key, ctx->dh);

#ifdef DEBUG
	  fprintf(stderr,"\nDH SHARED KEY\n");	
    print_dh_sharekey(ctx->share_key,ctx->share_keylen);
	  fprintf(stderr,"\n\n");	
#endif /*DEBUG*/

    DH_free(idh);
#ifdef DEBUG_BIO
    BIO_free(out);
#endif /*DEBUG_BIO*/
err:
    return(err_code);
}

/*
* Function: spkm3_gss_accept_sec_context()
*
* Description: 
* responds to init_sec_context queries, processes input_token
* returns output token
* 
* NOTE: spkm3's most basic form allows for no initiator or target credentials
* which is what is currently being coded.
*/ 

OM_uint32
spkm3_gss_accept_sec_context(
    void                    * context,
    OM_uint32               * minor_status,
    gss_ctx_id_t            * context_handle_P,
    gss_cred_id_t           verifier_cred_handle,
    gss_buffer_t            input_token,
    gss_channel_bindings_t  input_chan_bindings,
    gss_name_t              * src_name,
    gss_OID                 * mech_type,
    gss_buffer_t            output_token,
    OM_uint32               * ret_flags,
    OM_uint32               * time_rec,
    gss_cred_id_t           * delegated_cred_handle )
{
  spkm3_gss_ctx_id_desc  * ctx = NULL;
  OM_uint32              major_status = 0;
  int                    len;
  unsigned char          * ptr = 0,*pp =0;
  REQ_TOKEN              * req = NULL;
  REP_TI_TOKEN           * rep = NULL;
  int                    err_code;
  const gss_OID_desc     * mech_used = NULL;
  gss_buffer_desc        token;

#ifdef DEBUG
    fprintf(stderr, "spkm3_gss_accept_sec_context:\n") ;
#endif /* DEBUG */

    /* initialize variables */
    *minor_status = 0;
    if(src_name)
        *src_name = NULL;
    output_token->length = 0 ;
    output_token->value = NULL;
    ctx = *context_handle_P;

    /* spkm3: no credentials. */
    if(verifier_cred_handle != GSS_C_NO_CREDENTIAL)
        verifier_cred_handle = GSS_C_NO_CREDENTIAL;

    /* verify input token header integrity, including comparing the token 
    * mech_type with the mech argument. upon return ptr points past token
    * header, and stores remaining token length in len */

    ptr = (unsigned char *) input_token->value;
    if (!(err_code = g_verify_token_header((gss_OID) gss_mech_spkm3, &len, 
                &ptr, SPKM_TOK_CTX_REQ, input_token->length))) {
        *mech_type = gss_mech_spkm3;
        mech_used = gss_mech_spkm3;
    }
    if((ctx == (gss_ctx_id_t) GSS_C_NO_CONTEXT)){

        /* first time for this connection */

        /* decode token */
        pp = (char *)ptr;
        if((d2i_REQ_TOKEN(&req, &pp, (long)len)) == NULL)
            return GSS_S_DEFECTIVE_TOKEN;

#ifdef DEBUG
        /* print some of the token...*/
        print_req_token(req);
#endif /* DEBUG */

        /* check integrity of token */

        /* create and initialize context */
        major_status = spkm3_get_context(minor_status, (spkm3_ctx_id_t *)NULL,&ctx);
        err_code = spkm3_add_targname(ctx,req->reqcontents->targname->entries);

        /* receiver diffie-hellman */
        err_code = spkm3_receiver_dh(ctx,req);

        /* generate a REP_TI_TOKEN */
        *minor_status = spkm3_mk_dh_rep_ti(ctx,rep,req,mech_type,&token);
        if(*minor_status){
            major_status = GSS_S_FAILURE;
            goto err;
        }

        /* set all output parameters */
        ctx->spkm3_state = SPKM3_REQ_TI_TOKEN;
        *output_token = token;
        major_status = GSS_S_COMPLETE;

        /* fall through */
    } else {
      /* should not happen */
      fprintf(stderr,"ERROR accept_sec_context received NON NULL CONTEXT\n");
      major_status = GSS_S_FAILURE;
        /* fall through */
    }
err:
    return major_status;
}
