/*
 *  spkm3/init_sec_context.c
 *
 *  Copyright (c) 2001 The Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  Andy Adamson <andros@umich.edu>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <asn1.h>
#include <asn1_mac.h>
#include <ssl.h>
#include <gssapiP_generic.h>
#define OID_EQ 1
#include <gssapi_spkm3.h>

extern const gss_OID_desc * const gss_mech_spkm3;
extern const gss_OID_desc * const gss_ialg_nullmac;

/* hard coded null-mac ASN1_OBJECT */

ASN1_OBJECT null_mac_obj = {"NULL_MAC", "null-mac",0,6,"\052\006\001\005\003\003",9};


/* From kx509 source
 *
 * Function: req_fix_type()
 *
 * Description: Ensure acceptable type for supplied NID
 */

static int req_fix_type(nid,type)
int nid;
int *type;
{
    if (nid == NID_pkcs9_emailAddress)
        *type=V_ASN1_IA5STRING;
    if ((nid == NID_commonName) && (*type == V_ASN1_IA5STRING))
        *type=V_ASN1_T61STRING;
    if ((nid == NID_pkcs9_challengePassword) && (*type == V_ASN1_IA5STRING))
        *type=V_ASN1_T61STRING;
    if ((nid == NID_pkcs9_unstructuredName) && (*type == V_ASN1_T61STRING))
        return(0);
    if (nid == NID_pkcs9_unstructuredName)
        *type=V_ASN1_IA5STRING;
    return(1);
}

/*
* Function: add_DN_object()
*
* Description: add X509_NAME_ENTRY for given NID(with given value)
* to given DN (X509_NAME)
* 
* NOTE: from billdo's KCA code
*/

static int add_DN_object(X509_NAME *n,char *value,int nid)
{
  X509_NAME_ENTRY *ne=NULL;
  char buf[1024];
  int ret=0;
  int j;

  strcpy((char *)buf,value);

  j=ASN1_PRINTABLE_type(buf,-1);
  if (req_fix_type(nid,&j)
      && (ne=X509_NAME_ENTRY_create_by_NID(
          NULL,nid,j,buf, strlen((const char *)buf)))
      && X509_NAME_add_entry(n,ne,X509_NAME_entry_count(n),0))
    ret=1;

  if (ne != NULL)
    X509_NAME_ENTRY_free(ne);

  return(ret);
}

/*
* Function: spkm3_gss_to_x509_name()
*
* Description: creates an x509 subject name using a hardcoded
* preamble, and a gss_name_t name to construct the CN.
*
* NOTE: use the hard coded subject_dn declared in gssapi_spkm3.h
*/

OM_uint32
spkm3_gss_to_x509_name(
  spkm3_cn_service_t        target_name,
  X509_NAME                  *xname)
{
  OM_uint32 err_code = 0;
  UMICH_X509_DN *ent;
  int nid;

#ifdef DEBUG
    fprintf(stderr, "spkm3_gss_to_x509_name:\n") ;
#endif /* DEBUG */


    /* fill in <subject> with CA name */
    for (ent=subject_dn; ent->type; ent++)
    {
        nid=OBJ_txt2nid(ent->type);
        if (!add_DN_object(xname, ent->def_val, nid))
        {
        err_code = EINVAL;
        goto end;
        }
    }

    /*  add CN to subject */
    nid=OBJ_txt2nid("commonName");
    if (!add_DN_object(xname, target_name->data, nid))
    {
    err_code = EINVAL;
    goto end;
    }
end:
/*
    free(cname);
*/
    return err_code;
}

/*
* Function: spkm3_mk_dh_req()
*
* Description: creates the initiator side of the diffie-hellman
* request. constructs a wire-ready token to be passed to  
* the receiver gss_accept_sec_context(). 
*
* no initiator nor target certificate 
* dh inner context token format:
*  Tag [0] - [6]: SPKM-REQ   0;
*  requestToken
*    reg_contents            
*        tok-id							 0x0100
*        context-id          random-integer (fresh) passed to target
                                  which concatenates another random-#
																	to form the context-id which is
																	mapped to the context-handle
         pvno [0]            BIT STRINT protocol version # (1)
				 timestamp           UTCTime - OPTIONAL not used.
				 randSrc             random-integer
				 targ-name           Name -  of target machine.
				 src-name            spkm3 always anonymous
				 rep-data            context-data
				 validity [2]        Validity - OPTIONAL not used.
				 key-estb-set        Key-Estb-Algs - initiator 1/2 of the dh key
				 key-estb-req        BIT STRING  - OPTIONAL not used.
				 key-src-bing        OCTET STRING - OPTIONAL not used.
*    algId                   nullMAC
*    req-integrity           0 length bit string
*  certif-data - OPTIONAL not used.
*  auth-data - OPTIONAL not used.
*/

OM_uint32
spkm3_mk_dh_req (
    spkm3_gss_ctx_id_desc   *ctx,
		REQ_TOKEN *req,
    gss_OID                 mech_type,
    gss_buffer_t            token )
{
  OM_uint32 err_code = 0;
  unsigned char      *tok = 0,*ptr = 0;
  int                len=0,i=0,tlen=0;
	long int           randnum;
	struct timeval     tval;
	unsigned char      *s,*p;
  Req_contents       *reqcnt;
  ASN1_INTEGER       *ai = NULL;
  X509_ALGOR         *key_alg = NULL;
#ifdef DEBUG_BIO
  BIO                    *out = NULL;

    out=BIO_new(BIO_s_file());
    if (out == NULL) exit(1);
    BIO_set_fp(out,stderr,BIO_NOCLOSE);
#endif /*DEBUG_BIO*/


#ifdef DEBUG
    fprintf(stderr, "spkm3_mk_dh_req:\n") ;
#endif /* DEBUG */

    req = REQ_TOKEN_new();
		if (req == NULL) {
        err_code = ENOMEM;
        goto err; /* XXX make this real */
		}
  
    reqcnt = req->reqcontents;
	  /* set token id */
    if(!ASN1_INTEGER_set(reqcnt->tokid,0x0100)) goto err;

    /* context_id
		* XXX i know nothing about random numbers. should i set
		* a seed? if so, how? this needs to be a 'fresh' random number....
		*/
		gettimeofday(&tval,NULL);
		srand((unsigned int)tval.tv_sec);
    randnum =  rand();

#ifdef DEBUG
    fprintf(stderr, "spkm3_mk_dh_req: context_id rand1  %ld\n",randnum) ;
#endif /* DEBUG */

    /* XXX so, i der encode the random #, then set it as a BIT STRING 
    * is this correct? see crypto/asn1/x_pubkey.c for ASN1_BIT_STRING_set
    * example
    */
    if((ai = ASN1_INTEGER_new()) == NULL) {
        err_code = ENOMEM;
        goto err;
    }
    err_code = EINVAL; /*XXX*/
    if(!ASN1_INTEGER_set(ai,(int)randnum)) goto err;
    i = i2d_ASN1_INTEGER(ai,NULL);
		if ((s=(unsigned char *)Malloc(i+1)) == NULL) goto err;
		p=s;
		i2d_ASN1_INTEGER(ai,&p);
		if(!ASN1_BIT_STRING_set(reqcnt->contextid,s,i)) goto err;
    ASN1_INTEGER_free(ai);
		Free(s);
    err_code = 0;

    /* set protocol version, set bit 0 */
		ASN1_BIT_STRING_set_bit(reqcnt->pvno,0,1);
    
    /* skip timestamp */

    /* set randSrc */
    /* XXX get next random # from gettimeofday seed */
    randnum = rand();
    if((ai = ASN1_INTEGER_new()) == NULL) {
        err_code = ENOMEM;
        goto err;
    }
		err_code = EINVAL; /*XXX*/
    if(!ASN1_INTEGER_set(ai,(int)randnum)) goto err;
    i = i2d_ASN1_INTEGER(ai,NULL);
    if ((s=(unsigned char *)Malloc(i+1)) == NULL) goto err;
    p=s;
    i2d_ASN1_INTEGER(ai,&p);
    if(!ASN1_BIT_STRING_set(reqcnt->randSrc,s,i)) goto err;
    ASN1_INTEGER_free(ai);
    Free(s);
    err_code = 0;

    /* target name */
    err_code = spkm3_gss_to_x509_name(ctx->target_name,reqcnt->targname);
    if(err_code) goto err;

    /* source name  - this is OPTIONAL iff anonymous, which is always true
    * for spkm3. LIPKEY will pass a user name down to spkm3 for encrypting 
    */

    /* Context Data */
    err_code = spkm3_gss_context_data(reqcnt->reqdata);
    if(err_code) goto err;

    /* validity - OPTIONAL */

    /* key-estb-set 
    * the initiator 1/2 of the dh key info is split between the
    * keyestbset field which holds the parameters p & g 
    * and the keyestbreq which holds the initiators pubkey.
    */

    /*
    * Diffie-Hellman initiator.
    * Generate p and g. this can take a LONG time.... 
    */
    ctx->dh = DH_generate_parameters(64, DH_GENERATOR_5, NULL, NULL);

    /* Generate pub_key */
    DH_generate_key(ctx->dh);

#ifdef DEBUG_BIO
    BIO_puts(out,"\n\nINITIATOR pub_key =");
    BN_print(out,ctx->dh->pub_key);
    BIO_puts(out,"\n\n\n");
#endif /* DEBUG_BIO*/

    err_code = x509_algor_dh_set(&key_alg,NID_dhKeyAgreement,ctx->dh);
    if(err_code) goto err;
    if(!sk_push(reqcnt->keyestbset,(char *)key_alg)){
        err_code = ENOMEM;
        goto err;
    }

    /* key-estb-req - holds the diffie hellman initiator pub key*/
    i = i2d_DHPublicKey(ctx->dh,NULL); 
    if((s=(unsigned char *)malloc(i+1)) == NULL) {
        err_code = ENOMEM;
        goto err;
    }
    p=s;
    i2d_DHPublicKey(ctx->dh,&p);

    /* Req_contents_new() doesn't malloc OPTIONAL parameters */
    if(reqcnt->keyestbreq == NULL){
        if((reqcnt->keyestbreq = ASN1_BIT_STRING_new()) == NULL){
            err_code = ENOMEM;  /*XXXX*/ 
            Free(s);
            goto err;
        }
    }
    if (!ASN1_BIT_STRING_set(reqcnt->keyestbreq,s,i)) { 
        err_code = ENOMEM;  /*XXXX*/ 
        goto err;
    }
    Free(s);
    
    /* key-src-bind OPTIONAL*/

    /* algId */
    err_code = x509_algid_set(req->algId);
    if(err_code) goto err;
    
    /* req->reqintegrity: do nothing, null-mac -> NULL BIT_STRING */
    
    /* calculate the length of the TOKEN */
    len = i2d_REQ_TOKEN(req,NULL);
    
    /* token stuff from krb5/make_ap_req_v1()*/
    /* build up the token */

    /* allocate space for the token */
    tlen = g_token_size((gss_OID) mech_type, len);

    if ((tok = (unsigned char *) xmalloc(tlen)) == NULL) {
        err_code = ENOMEM;
        goto cleanup;
    }

    /* fill in the buffer */
    ptr = tok;

    g_make_token_header((gss_OID) mech_type, len,
            &ptr, SPKM_TOK_CTX_REQ);

    /* krb5 does a TWRITE_STR here which copies the asn1 encoded 
    * req data into tok, and advances ptr. we call i2d which encodes 
    * the request data into tok and advances the ptr.
    */
    len  = i2d_REQ_TOKEN(req, &ptr);


    token->length = tlen;
    token->value = (void *) tok;
cleanup:
    REQ_TOKEN_free(req);
#ifdef DEBUG_BIO
    BIO_free(out);
#endif /*DEBUG_BIO*/
err:
    return err_code;
}


/*
* Function: spkm3_gss_init_sec_context()
*
* Description: 
* called by initiator in a loop, sending output_token to target,
* and processing returned input_token until the protocol 
* is done.
* 
* NOTE: according to rfc 2847, spkm3_gss_init_sec_context will be called
* from the LIPKEY mechanism's gss_init_sec_context(). 
*
* NOTE: spkm3's most basic form allows for no initiator or target credentials
* which is what is currently being coded.
*/ 

OM_uint32
spkm3_gss_init_sec_context (
    void  *                        cont,
    OM_uint32 *                    minor_status ,
    const gss_cred_id_t            initiator_cred_handle ,
    gss_ctx_id_t *                 context_handle_P ,
    gss_name_t                     target_name ,
    gss_OID                        mech_type ,
    OM_uint32                      req_flags ,
    OM_uint32                      time_req ,
    gss_channel_bindings_t         input_chan_bindings ,
    gss_buffer_t                   input_token ,
    gss_OID *                      actual_mech_type ,
    gss_buffer_t                   output_token ,
    OM_uint32 *                    ret_flags ,
    OM_uint32 *                    time_rec )
{

  spkm3_gss_ctx_id_desc  *ctx = NULL;
  OM_uint32              major_status,err;
  REQ_TOKEN              *req=NULL;
  REP_TI_TOKEN           *rep=NULL;
  gss_buffer_desc        token;
  unsigned char          *ptr=0,*pp=0;
  int                    len; 
  int                    j = 0,key_len=0;
  DH                     *rdh=NULL;
  unsigned char          *p;
#ifdef DEBUG
  BIO                    *out = NULL;

    out=BIO_new(BIO_s_file());
    if (out == NULL) exit(1);
    BIO_set_fp(out,stderr,BIO_NOCLOSE);
#endif /* DEBUG */

#ifdef DEBUG
    fprintf(stderr, "spkm3_gss_init_sec_context:\n") ;
#endif /*DEBUG*/

/* initialize variables */
    major_status = GSS_S_FAILURE;
    err = 0;
    *minor_status = 0;
    if(output_token){
        output_token->length = 0 ;
        output_token->value = NULL;
    }

    /* verify target name is valid and useable.*/ 
    /* check for spkm3 compliance */

    /* spkm3 - client has no credential */
    if(initiator_cred_handle != GSS_C_NO_CREDENTIAL)
        return GSS_S_FAILURE;
    
    /* spkm3 - no mutual auth */
    if(req_flags & GSS_C_MUTUAL_FLAG)
        return GSS_S_FAILURE;
	
    /* spkm3 - anon flag MUST be true ;) */
    if(!(req_flags & GSS_C_ANON_FLAG))
        return GSS_S_FAILURE;

    /* spkm3 - mech MUST be OID for spkm3  NOTE: g_OID_equal part of Krb5 glue*/
    if(!(g_OID_equal(mech_type, gss_mech_spkm3)))
        return GSS_S_BAD_MECH;


    if ((*context_handle_P == (gss_ctx_id_t) GSS_C_NO_CONTEXT)) {

        /* first time for this connection */

        /* spkm3 - input_token MUST be null */
        if(input_token != GSS_C_NO_BUFFER && input_token->length !=0)
            return GSS_S_DEFECTIVE_TOKEN;

        major_status = spkm3_gss_init_context(minor_status,&ctx);
        if(GSS_ERROR(major_status))
            return GSS_S_FAILURE;
    
        major_status = spkm3_construct_target_name(ctx,(spkm3_cn_service_t)target_name);
#ifdef DEBUG
        print_cn_service(ctx->target_name);
#endif /*DEBUG*/
        
        *context_handle_P = ctx;
        ctx->req_flags = req_flags;

        if (actual_mech_type != NULL) {
            *actual_mech_type = (gss_OID) gss_mech_spkm3;
        }

        if (ret_flags != NULL) {
           *ret_flags = 0 ;
        }

        if (time_rec != NULL) {
            *time_rec = GSS_C_INDEFINITE;
        }
        *minor_status = spkm3_mk_dh_req(ctx,req,mech_type,&token);
        if(*minor_status)
            major_status = GSS_S_FAILURE;
        else major_status = GSS_S_CONTINUE_NEEDED;

        *output_token = token;
    } /* done first time */
    else {
        fprintf(stderr,"SPKM3_REQ_TI_TOKEN\n");
        if(!input_token) return GSS_S_FAILURE; 
        ctx = *context_handle_P;

        /* check integrity of token */

        /* verify input token header integrity, including comparing the 
        * token mech_type
        * with the mech argument. upon return ptr points past token
        * header, and stores remaining token length in len 
        */

        ptr = (unsigned char *) input_token->value;
        if (!(*minor_status = g_verify_token_header((gss_OID) gss_mech_spkm3, &len,
                &ptr, SPKM_TOK_CTX_REP, input_token->length))) {
            if (actual_mech_type != NULL)
                *actual_mech_type = gss_mech_spkm3;
        }
    
        /* decode token */
        pp=ptr;
        if((d2i_REP_TI_TOKEN(&rep, &pp, (long)len)) == NULL)
            return GSS_S_DEFECTIVE_TOKEN;
#ifdef DEBUG
        /* print some of the token...*/
        print_rep_ti_token(rep);
#endif /* DEBUG */
        /* grab receiver pub key from the wire*/  
        p = rep->repticontents->keyestbstr->data;
        j = rep->repticontents->keyestbstr->length;
        if((rdh = d2i_DHPublicKey(NULL, &p, (long)j)) == NULL) {
            *minor_status = ENOMEM; /*XXX*/
            return GSS_S_FAILURE;
        }

#ifdef DEBUG_BIO
        BIO_puts(out,"\n\nRECEIVOR pub_key =");
        BN_print(out,rdh->pub_key);
        BIO_puts(out,"\n\n");
#endif /*DEBUG_BIO*/
        /* call DH_compute_key to create shared key*/
        key_len = DH_size(ctx->dh);
        ctx->share_key = (unsigned char *)malloc(key_len*sizeof(unsigned char));
        ctx->share_keylen = DH_compute_key(ctx->share_key,rdh->pub_key,ctx->dh);
#ifdef DEBUG
        fprintf(stderr,"\nDH SHARED KEY\n");
        print_dh_sharekey(ctx->share_key,ctx->share_keylen);
        fprintf(stderr,"\n\n");
#endif /*DEBUG*/
          
        major_status =  GSS_S_COMPLETE;
    }
    return major_status; 
}

/*
* Function:  spkm3_gss_context_data
* 
* Description: 
* called by spkm3_mk_dh_req to pack the Context Data int the 
* REQ_TOKEN.
* 
*/

OM_uint32
spkm3_gss_context_data(
  Context_Data *reqdata)
{
  X509_ALGOR *ciph = NULL, *intg_alg = NULL, *owf_alg = NULL;
  OM_uint32 ret = 0;

#ifdef DEBUG
    fprintf(stderr, "spkm3_gss_context_data\n") ;
#endif /* DEBUG */

    /* channelId - OPTIONAL - not used in first pass coding */

    /* seq-number  - OPTIONAL - not used in first pass coding */

    /* options - the set of features that will be active on the context.
    * see rfc 1508
    * (0) delegation-state - delegation_req_flag set  
    * (1) mutual-state - this is not set, no mutual authentication in spkm3.
    * (2) replay-det-state -  replay_det_flag set
    * (3) sequence-state - 
    * (4) conf-avail - per-message confidentialiy available
    * (5) integ-avail - per-message integrity services available
    * (6) target-certif-data-required - MUST be set for spkm3 
    * first pass coding: just set bit 6.
    */

    /* should already be allocated... */
    if(reqdata->options == NULL)  {
          if ((reqdata->options = ASN1_BIT_STRING_new()) == NULL) return ENOMEM;
    }

    ASN1_BIT_STRING_set_bit(reqdata->options,6,1);

    /* Conf-Algs 
    * add cast5-cbc as per rfc 2847
    *
    * this is used to encode the initialization vector once the context
    * is set up
    * ret=EVP_CIPHER_set_asn1_iv(cipher,a->parameter);
    * 
    * XXXX x509_algor_list_set creates a new X509_ALGOR.
    * when is it freed, in Conf_Algs_free???
    */
    ret = x509_algor_list_set(&ciph,(EVP_cast5_cbc())->nid);
    reqdata->confalg->tag = 0;
    ret = Conf_Alg_add_entry(reqdata->confalg,ciph);

    /* Intg-Algs
    * add md5WithRSAEncryption as RECOMMENDED in rfc2847
    */
    ret = x509_algor_list_set(&intg_alg,(EVP_md5())->type);
    if(!sk_push(reqdata->intgalg,(char *)intg_alg))
        return -1;

    /* OWF-Algs 
    * add id-sha1 as per rfc2847
    */
    ret = x509_algor_list_set(&owf_alg,(EVP_sha1())->type);
    if(!sk_push(reqdata->owfalg,(char *)owf_alg))
        return -1;

    return 0;
}

/*
* Function:  Conf_Algs_add_entry
*
* Description: add an X509_ALGOR to the Conf_Alg stack
*
*/

OM_uint32
Conf_Alg_add_entry (
  Conf_Algs *conf_alg,
  X509_ALGOR *alg)
{
    if((conf_alg == NULL) || (alg == NULL)) 
        return EINVAL;

    if(conf_alg->d.algs == NULL) 
        conf_alg->d.algs = sk_new_null();
    if(conf_alg->d.algs == NULL) return ENOMEM;

    if(!sk_push(conf_alg->d.algs,(char *)alg))
        return ENOMEM;

    return 0;
}

/*
* Function:  x509_algor_list_set
*
* Description: set the AlgorithmIdentifier (X509_ALGOR) entry
* with nid and null parameter.
*
* Returns: 1 on success, 0 on failure
*/

OM_uint32
x509_algor_list_set (
  X509_ALGOR **x,
  int nid)
{
  X509_ALGOR *a;
  ASN1_OBJECT *o;
  int ret = 0;

    if((a=X509_ALGOR_new()) == NULL) goto err;

    /* set the algorithm id */
    if ((o=OBJ_nid2obj(nid)) == NULL) goto err;
    ASN1_OBJECT_free(a->algorithm);
    a->algorithm=o;
    /* this is a list, so null parameters */
    ASN1_TYPE_free(a->parameter);
    a->parameter=ASN1_TYPE_new();
    a->parameter->type=V_ASN1_NULL;
    
err:
   *x = a;
   return ret;
}

/*
* Function:  x509_algid_set
*
* Description:  stuff the algId with the NULL-MAC Ialg
* as per rfc2847. NOTE: null-mac is NOT known to SSLeay.
* i've been able to add it to openssl.....
*
* the hard coded null_mac_obj is a first try - there is no nid
* which could be fatal....
*
* Returns: 0 on success
*/

OM_uint32
x509_algid_set (
  X509_ALGOR *x)
{
    /* set the algorithm id */
    ASN1_OBJECT_free(x->algorithm);
    x->algorithm=&null_mac_obj;

    ASN1_TYPE_free(x->parameter);
    x->parameter=ASN1_TYPE_new();
    x->parameter->type=V_ASN1_NULL;
    
   return 0;
}


/*
* Function:  x509_algor_dh_set
*
* Description:  stuff the key_alg with diffie hellmen
* initiator 1/2 of key.
*
* Returns: 0 on success
*/

OM_uint32
x509_algor_dh_set (
  X509_ALGOR **x,
  int nid,
  DH *dh)
{
  X509_ALGOR *a;
  ASN1_OBJECT *o;
  int i;
  unsigned char *pp,*p;

    if((a=X509_ALGOR_new()) == NULL) goto err;

    /* set the algorithm id */
    if ((o=OBJ_nid2obj(nid)) == NULL) goto err;
    ASN1_OBJECT_free(a->algorithm);
    a->algorithm=o;

    ASN1_TYPE_free(a->parameter);
    i=i2d_DHparams(dh,NULL);
    p=(unsigned char *)Malloc(i);
    pp=p; 
    i2d_DHparams(dh,&pp);
    a->parameter=ASN1_TYPE_new();
    a->parameter->type=V_ASN1_SEQUENCE;
    a->parameter->value.sequence=ASN1_STRING_new();
    ASN1_STRING_set(a->parameter->value.sequence,p,i);
    Free(p);
    
err:
   *x = a;
   return 0;
}
